from collections import OrderedDict, Counter
from pathlib import Path
import time
import requests
from typing import Iterable, Dict, Union
import logging

from ankipan import Reader, Scraper, File, ANKIPAN_DB_ADDR

logger = logging.getLogger(__name__)

class Deck:
    def __init__(self,
                 source_lang: str,
                 native_lang: str,
                 source_words = None,
                 learning_collection_words: set = None,
                 known_collection_words: set = None,
                 sentence_sources: set = None):
        self.source_lang = source_lang
        self.reader = Reader(source_lang)
        self.scraper = Scraper(source_lang, native_lang)
        self.source_words = source_words if source_words else Counter()
        self.ignored_words = set()
        self.learning_collection_words = learning_collection_words if learning_collection_words else set()
        self.known_collection_words = known_collection_words if known_collection_words else set()
        self.skip_words = set()
        self.percentiles = {}
        self.scraped_definitions = {}
        self.sentence_sources = list(sentence_sources) if sentence_sources else []
        self._word_percentiles = None
        self._sorted_words = None
        self._domain_occurrences = None

    @property
    def all_words(self):
        return {word for word in self.source_words}

    @property
    def new_words(self):
        return Counter({word: count for word, count in self.source_words.items() if word not in self.learning_collection_words and word not in self.known_collection_words and word not in self.ignored_words and word not in self.skip_words})

    @property
    def learning_words(self):
        """Get words that are already being learned in collection"""
        return Counter({word: count for word, count in self.source_words.items() if word in self.learning_collection_words})

    @property
    def known_words(self):
        """Get words that are already specified as known word in collection"""
        return Counter({word: count for word, count in self.source_words.items() if word in self.known_collection_words})

    @property
    def domain_occurrences(self):
        if self._domain_occurrences is None:
            try:
                self._domain_occurrences = requests.post(f'{ANKIPAN_DB_ADDR}/api/corpus_wordcounts',
                                                         json={'lang': self.source_lang,
                                                               'words': list(self.source_words.keys()),
                                                               'subdomains': self.sentence_sources}).json()
            except:
                logger.error('Unable to reach remote server, not fetching domain occurrences')
                return {}
        return self._domain_occurrences

    @property
    def word_percentiles(self):
        if not self._word_percentiles:
            domains = {'source': self.source_words} | self.domain_occurrences
            self._word_percentiles = self.calculate_percentiles(self.new_words.keys(), domains)
        return self._word_percentiles

    @property
    def sorted_words(self):
        # TODO: does not yet account for when words are literally added or removed
        if not self._sorted_words:
            min_percentiles = {
                word: self.word_percentiles[word].get('source', 1) for word in self.word_percentiles
            }
            self._sorted_words = sorted(min_percentiles.keys(), key=lambda w: min_percentiles[w])
        return self._sorted_words

    def select_new_words(self, n_words=None):
        """
        Displays words in an ipysheet 'spreadsheet'. Each row has 'Skip', 'Known',
        'Word', and dynamically generated percentile columns.
        """
        import ipysheet
        from ipysheet import sheet, column, cell
        from ipywidgets import Button, VBox, Layout
        from IPython.display import display
        from IPython import get_ipython
        from functools import partial
        import ipywidgets as widgets

        if not get_ipython():
            raise NotImplementedError("Spreadsheet UI only available in Jupyter.")
        words_to_show = self.sorted_words if n_words is None else self.sorted_words[:n_words]
        all_domains   = sorted({d for v in self.word_percentiles.values() for d in v})

        n_rows = len(words_to_show)
        s      = sheet(rows=n_rows, columns=3 + len(all_domains))
        s.column_headers = ["Skip", "Known", "Word"] + all_domains

        col_sel   = column(0, [False]*n_rows, type="checkbox", label="Skip",  sheet=s)
        col_known = column(1, [False]*n_rows, type="checkbox", label="Known", sheet=s)
        col_word  = column(2, words_to_show,  type="text",     label="Word",  sheet=s)

        def bg(val: float | None) -> str:
            if val is None:
                return "#ffffff"
            if val < 0.10:
                return "#c6f7c6"
            if val < 0.25:
                return "#fff7b5"
            return "#f7c6c6"
        for col_idx, domain in enumerate(all_domains, start=3):
            for row_idx, word in enumerate(words_to_show):
                val = self.word_percentiles.get(word, {}).get(domain)
                txt = "" if val is None else f"{val:.4f}"
                ipysheet.cell(
                    row=row_idx,
                    column=col_idx,
                    value=txt,
                    sheet=s,
                    type="text",
                    background_color=bg(val),
                    layout=Layout(width="70px")
                )

        btn  = widgets.Button(description="Set unknown words")
        out  = widgets.Output()
        def on_click(b):
            with out:
                out.clear_output()
                err_mask   = col_sel.value
                known_mask = col_known.value
                words      = col_word.value
                skip_words  = [w for w, m in zip(words, err_mask)   if m]
                known      = [w for w, m in zip(words, known_mask) if m]
                unknown = set(self.source_words) - set(skip_words) - set(known)
                self.set_new_words(unknown, skip_words)
                print("✅ Skip Words:", skip_words[:10], "...")
                print("✅ Known    :", known[:10], "...")
                print("➡️ New  :", sorted(unknown)[:10], "...")
        btn.on_click(on_click)
        display(widgets.VBox([s, btn, out]))

    def set_new_words(self, new_words: Iterable, skip_words, ignore_unknown=False):
        for word in self.source_words.keys():
            if word in new_words:
                self.known_collection_words.discard(word)
                self.learning_collection_words.discard(word)
            elif word in skip_words:
                self.skip_words.add(word)
            elif word not in self.known_collection_words and word not in self.learning_collection_words:
                self.known_collection_words.add(word)
        for word in new_words:
            if word not in self.all_words:
                print(f'Warning: word {word} not in source words, ' +
                      f'{"ignoring" if ignore_unknown else "adding with occurrence 1"}')
                if not ignore_unknown:
                    self.source_words[word] = 1

    def set_ignored_words(self, ignored_words: Iterable):
        for word in ignored_words:
            self.ignored_words.add(word)
            if word in self.known_collection_words:
                self.known_collection_words.remove(word)
            if word in self.learning_collection_words:
                self.learning_collection_words.remove(word)

    def calculate_percentiles(self, words: Iterable[str], domains: OrderedDict[str, Counter]) -> Dict[str, Dict[str, float]]:
        if not hasattr(self, 'percentiles'):
            self.percentiles = {}
        percentiles = {word: {} for word in words}
        for word in words:
            if word not in self.percentiles:
                for domain_name, word_count_dict in domains.items():
                    counter = Counter(word_count_dict)
                    if word in counter:
                        common = [word for word, count in counter.most_common()]
                        percentile = common.index(word) / len(counter) if counter[word] != 1 else 1.0
                        percentiles[word][domain_name] = percentile
                    else:
                        percentiles[word][domain_name] = None
                self.percentiles[word] = percentiles[word]
            else:
                percentiles[word] = self.percentiles[word]
        return percentiles

    def __repr__(self):
        return str(self)

    def __str__(self):
        if not any([self.new_words, self.learning_words, self.known_words]):
            return 'No words in collection'
        categories = [f'New Words ({len(self.new_words)})',
                      f'Learning Words ({len(self.learning_words)})',
                      f'Known Words ({len(self.known_words)})']
        domains = {'source': self.source_words} | self.domain_occurrences
        lines = []
        col_widths = [max(len(domain_name) + 8, 10) for domain_name in domains]  # Adjust width for "source" and "global"
        word_width = max(14, max(len(word) for category in [self.new_words, self.learning_words, self.known_words] for word in category.keys()) + 2)
        header_line = ' '.join(['word'.ljust(word_width)] + [f"{header}".ljust(width) for header, width in zip(domains.keys(), col_widths)])
        lines.append(' | '.join([f'{category:<{sum(col_widths) + word_width + 2}}' for category in categories]))
        lines.append("_" * (sum(col_widths) + word_width + 3 * len(categories) + 2 * (len(categories)-1)))
        lines.append(' | '.join([f'{header_line:<{sum(col_widths) + word_width + 2}}' for _ in categories]))
        lines.append('')
        data = {}
        for category, counter in zip(categories, [self.new_words, self.learning_words, self.known_words]):
            # TODO replace with properties
            categorypercentiles = self.calculate_percentiles(counter.keys(), domains)
            sorted_words = sorted(categorypercentiles.keys(), key=lambda word: categorypercentiles[word].get('source', float('inf')))
            data[category] = {word: categorypercentiles[word] for word in sorted_words}
        max_length = max(len(data[category]) for category in data)
        for i in range(max_length):
            row_parts = []
            for category in categories:
                words = list(data[category].keys())
                if i < len(words):
                    word = words[i]
                    row = word.ljust(word_width)
                    for domain_name in domains:
                        percentile = data[category][word].get(domain_name)
                        if percentile is None:
                            text = "N/A"
                        elif percentile == 1.0:
                            text = "100.00%"
                        else:
                            text = "{:.2%}".format(percentile)
                        row += text.ljust(col_widths[list(domains.keys()).index(domain_name)])
                else:
                    row = " " * word_width
                row_parts.append(f'{row:<{sum(col_widths) + word_width + 2}}')
            lines.append(" | ".join(row_parts))
        return "\n".join(lines)

    def add(self,
            path: Union[str, Path] = None,
            *,
            string: str = None,
            word_counts: Union[Dict[str, int], Counter] = None):
        """
        Add words from file(s) to word collection

        Parameters
        ----------
        path: path to file(s)
        string (optional): parse string instead of file, only valid if no file is specified

        """
        if path and string:
            raise RuntimeError('Please only supply either a path or a string.')
        elif word_counts is not None:
            if not (isinstance(word_counts, dict) or isinstance(words, Counter)):
                raise RuntimeError(f'Deck requires Dict- or Counter like datastructure to update, received {type(word_counts)}:\n  {word_counts}')
        else:
            if string is not None:
                files = [File(self.source_lang, string,)]
            else:
                file_paths = self.reader.collect_file_paths(path)
                files = self.reader.open_files(file_paths)
            self.reader.process_files(files)
            word_counts = Counter()
            for file in files:
                word_counts.update(file.word_count)
        self.source_words.update(word_counts)

    def remove(self, words: Union[str, Iterable]):
        """
        Remove words from word collection

        Parameters
        ----------
        words: words to remove

        """
        if isinstance(words, str):
            words = [words]
        elif not (isinstance(words, list) or isinstance(words, set)):
            raise RuntimeError('Only string or list allowed in collection.remove command')
        for word in words:
            if word not in self.source_words: raise RuntimeError(f'Word "{word}" is not part of this wordcollection, abort.')
        [self.source_words.pop(word) for word in words]

    def remove_range(self, lower: int, upper: int):
        self.remove([word for word, count in self.source_words.items() if count >= lower and count < upper])



import requests
import json
import logging

from typing import Tuple, Iterable

logger = logging.getLogger(__name__)

class AnkiManager:
    def __init__(self, address='http://localhost:8765'):
        self.address = address

    def _invoke(self, action, **params):
        request_json = json.dumps({'action': action, 'params': params, 'version': 6})
        response = requests.post(self.address, data=request_json)
        return response.json()

    def create_deck(self, deck_name):
        return self._invoke('createDeck', deck=deck_name)

    def delete_deck(self, deck_name):
        return self._invoke('deleteDeck', decks=[deck_name])

    def get_deck_names(self):
        return self._invoke('deckNames')['result']

    def add_note(self, anki_deck_name, front, back, model_name="Basic"):
        note = {
            "deckName": anki_deck_name,
            "modelName": model_name,
            "fields": {
                "Front": front,
                "Back": back
            },
            "options": {"allowDuplicate": False},
            "tags": []
        }
        return self._invoke('addNote', note=note)

    def add_notes(self, anki_deck_name, new_cards: Iterable['Card'], section_names):
        notes_to_add = [{
            "deckName": anki_deck_name,
            "modelName": "Basic",
            "fields": {
                "Front": card.generate_front(),
                "Back": card.generate_back(section_names)
            },
            "options": {"allowDuplicate": False},
            "tags": []
        } for card in new_cards]
        res = self._invoke('addNotes', notes=notes_to_add)
        # TODO: Warning: Notes already exist in other deck, not adding here.
        if res['error'] is not None:
            if 'cannot create note because it is a duplicate' in res['error']:
                logger.error('Unable to sync with Ankiconnect because the anki collection that is currently open in the desktop app already contains the same cards from a different deck. Anki does not allow duplicate card frontsides, so users must use the same anki collection/profile for each ankipan collection.')
            raise RuntimeError(f"All anki add_notes errors: {res['error']}")
        return res['result'] if isinstance(res['result'], list) else [res['result']]

    def find_notes(self, query):
        return self._invoke('findNotes', query=query)['result']

    def get_notes_info(self, note_ids):
        return self._invoke('notesInfo', notes=note_ids)['result']

    def update_note(self, note_id, card, section_names):
        note = {"id": note_id, "fields": {"Front": card.generate_front(), "Back": card.generate_back(section_names)}}
        return self._invoke('updateNoteFields', note=note)

    def delete_notes(self, note_ids):
        return self._invoke('deleteNotes', notes=note_ids)

    def sync(self):
        return self._invoke('sync')

    def get_notes_for_anki_deck(self, anki_deck_name):
        note_ids = self.find_notes(f'"deck: {anki_deck_name}"')
        return self.get_notes_info(note_ids)

    def sync_deck(self, deck_name, cards, section_names, overwrite = False):
        existing_notes = self.get_notes_for_anki_deck(deck_name)
        if not existing_notes and deck_name not in self.get_deck_names():
            self.create_deck(deck_name)
        existing_anki_ids = {note['noteId'] for note in existing_notes}
        if overwrite:
            deleted_notes = self.delete_notes(list(existing_anki_ids))
        cards_to_update = [card for card in cards if card.anki_id and card.anki_id in existing_anki_ids]
        new_cards = [card for card in cards if not card.anki_id or card.anki_id not in existing_anki_ids]
        if cards_to_update:
            for card in cards_to_update:
                self.update_note(card.anki_id, card, section_names)
        if new_cards:
            anki_ids = self.add_notes(deck_name, new_cards, section_names)
            for card, anki_id in zip(cards, anki_ids):
                if anki_id and anki_id != 'error':
                    card.anki_id = anki_id

    # TODO
    # def sync_images(self, images_path: Path):
    #     def send_request(action, **params):
    #         response = requests.post('http://localhost:8765', json={"action": action, "version": 6, "params": params})
    #         response.raise_for_status()  # This will raise an HTTPError if the HTTP request returned an unsuccessful status code.
    #         return response.json()
    #     anki_media_root = send_request('getMediaDirPath')['result']
    #     anki_media_files = set(os.listdir(anki_media_root))
    #     deck_images = set(os.listdir(images_path))
    #     for img in anki_media_files-deck_images:
    #         os.remove(Path(anki_media_root, img))
    #     for img in deck_images-anki_media_files:
    #         shutil.copy(Path(images_path, img), anki_media_root)



from pathlib import Path
import yaml

PROJECT_ROOT = Path(__file__).parent.parent
ANKIPAN_ROOT = PROJECT_ROOT / 'ankipan'
HTML_TEMPLATE_DIR = ANKIPAN_ROOT / 'html_templates'
ANKI_COLLECTIONS_DIR = PROJECT_ROOT / 'anki_collections'

DATA_DIR = PROJECT_ROOT / '.data'
USER_DATA_FILE = PROJECT_ROOT / '.user_data.yaml'
if USER_DATA_FILE.exists():
    with open(USER_DATA_FILE, 'r') as f:
        USER_DATA = yaml.safe_load(f)
else:
    USER_DATA = {}

# TODO: run server remotely and change this
ANKIPAN_DB_ADDR = 'http://localhost:5000'

from .util import *
from .anki_manager import AnkiManager
from .reader import Reader, File
from .scraper import Scraper, CardSection
from .card import Card
from .deck import Deck
from .collection import Collection



import numpy as np
from collections import Counter
import pickle
from ankipan import PROJECT_ROOT
from pathlib import Path
from collections import Counter
import json

from typing import Any

def load_json(path: Path) -> dict[str, Any]:
    if path.exists():
        with path.open("r", encoding="utf-8") as fh:
            return json.load(fh)
    else:
        return {}

def save_json(path: Path, obj: dict[str, Any]) -> None:
    with path.open("w", encoding="utf-8") as fh:
        json.dump(obj, fh, indent=2, ensure_ascii=False)

def estimate_proficiency(word_counter: Counter, simulated_responses=None):
    """
    Very minimal method to estimate a user's initial proficiency level:
    Pick large corpus, sort by frequency, assume that less frequent words are less likely to be known.
    Query for words of different frequency, estimate ballpark interval of proficiency level.

    Returns
    -------

    A list of words that the use is likely to already know, alongside the percentile value

    """
    # TODO: Currently, we just create a zipf curve from our corpus, and estimating where a reasonable cut off index could be.
    # A more sophisticated method would detect areas the user is knowledgeable at, and areas where he/she is lacking.

    words = list(word_counter.keys())
    frequencies = list(word_counter.values())

    # Sort words by frequency (most frequent to least frequent)
    sorted_indices = np.argsort(frequencies)[::-1]
    sorted_words = [words[i] for i in sorted_indices]

    # Create a dictionary for fast index lookup
    word_to_index = {word: idx for idx, word in enumerate(words)}

    def initialize_probabilities(words):
        probabilities = np.ones(len(words)) / len(words)
        return probabilities

    def update_probabilities(words, probabilities, group, response):
        indices = [word_to_index[word] for word in group]
        known = np.array(response)

        for i, word in enumerate(group):
            index = indices[i]
            if known[i]:
                probabilities[index] *= 1.5
            else:
                probabilities[index] *= 0.7

        probabilities /= np.sum(probabilities)
        return probabilities

    def estimate_level(probabilities):
        cumulative = np.cumsum(probabilities)
        level = np.argmax(cumulative >= 0.5)
        return level

    def select_words(percentiles, words, words_per_question):
        selected_words = []
        for percentile in percentiles:
            start_idx = max(0, int(percentile * len(words) / 100))
            end_idx = min(len(words), start_idx + max(1, len(words) // 100))
            if end_idx > start_idx:
                selected_words.extend(np.random.choice(words[start_idx:end_idx], 1, replace=False))
        return selected_words[:words_per_question]

    initial_percentiles = [20, 40, 60, 80]
    probabilities = initialize_probabilities(words)
    num_questions = 5
    words_per_question = 5

    for question_num in range(num_questions):
        group = select_words(initial_percentiles, sorted_words, words_per_question)
        group.sort()
        percentiles = [
            (sorted_words.index(word) / len(words)) * 100
            for word in group
        ]
        if simulated_responses:
            response = simulated_responses[question_num]
        else:
            response = input(f"{list(zip(group, percentiles))} Do you know these words? Enter 'y' or 'n' for each, separated by space: ").split()
            if len(response) != words_per_question or any(r not in ['y', 'n'] for r in response):
                print(f"Please enter exactly {words_per_question} y/n responses.")
                continue

        response = [r.lower() == 'y' for r in response]
        probabilities = update_probabilities(words, probabilities, group, response)

        if all(response):
            initial_percentiles = [min(p + 10, 100) for p in initial_percentiles]  # Move to harder words
        elif not any(response):
            initial_percentiles = [max(p - 10, 0) for p in initial_percentiles]  # Move to easier words
        else:
            lower_bound = min(p for p, r in zip(percentiles, response) if r)
            upper_bound = max(p for p, r in zip(percentiles, response) if r)
            mid_point = (lower_bound + upper_bound) / 2
            initial_percentiles = [max(mid_point - 10, 0), mid_point, min(mid_point + 10, 100)]

    estimated_level = estimate_level(probabilities) / 3
    percentile = estimated_level / len(words) * 100
    return estimated_level, percentile

if __name__ == "__main__":
    with open(f'{PROJECT_ROOT}/corpus_resources/de.pkl', "rb") as f:
        corpus_occurrences_ = pickle.load(f)
    estimate_proficiency(corpus_occurrences_)




import re
import pysubs2
import os
from collections import Counter, OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm.notebook import tqdm
from pathlib import Path
from typing import Union, Dict, Any, Iterable, List, Optional, Tuple
import pickle
import logging
import chardet
from intervaltree import Interval, IntervalTree
import re
from bs4 import BeautifulSoup
import unicodedata
from langdetect import detect
from langdetect.lang_detect_exception import LangDetectException

from ankipan import PROJECT_ROOT

logger = logging.getLogger(__name__)

hanta_map = {
    'de': 'morphmodel_ger.pgz'
}
_nlp_hanta = None

stanza_map = {
    'jp': 'ja'
}

pattern = re.compile(
    r'[^'
    r'\s'
    r'\u3040-\u309F'  # Hiragana
    r'\u30A0-\u30FF'  # Katakana
    r'\u4E00-\u9FFF'  # Kanji
    r'\u3000-\u303F'  # Japanese punctuation
    r'\uFF10-\uFF19'  # Full-width digits
    r'\uFF21-\uFF3A'  # Full-width uppercase letters
    r'\uFF41-\uFF5A'  # Full-width lowercase letters
    r'a-zA-Z0-9'      # ASCII alphanumerics
    r'.,!?。、！？'    # Common punctuation marks
    r'\u00C0-\u00FF'  # Latin-1 Supplement (e.g., ä, ö, ü, ß, ç, é)
    r'\u0100-\u017F'  # Latin Extended-A (e.g., Œ, œ, Ā, Ă, Ą)
    r']',
)

class Reader():
    def __init__(self, source_lang):
        self.source_lang = source_lang
        self._nlp_stanza = None
        self.corpus_occurrences_ = None
        self.torch_ = None
    @property
    def torch(self):
        if not self.torch_:
            import torch
            self.torch_ = torch
        return self.torch_

    @property
    def nlp_stanza(self):
        if self._nlp_stanza is None:
            import stanza
            stanza_lang_name = stanza_map.get(self.source_lang, self.source_lang)
            try:
                self._nlp_stanza = stanza.Pipeline(stanza_lang_name, download_method=None, processors='tokenize,pos,lemma', use_gpu=True, tokenize_mode='A')
            except (stanza.pipeline.core.LanguageNotDownloadedError, stanza.resources.common.ResourcesFileNotFoundError, FileNotFoundError):
                print(f'Downloading stanza model for {stanza_lang_name}...')
                stanza.download(stanza_lang_name)
                print(f'Stanza model for {stanza_lang_name} downloaded')
                self._nlp_stanza = stanza.Pipeline(stanza_lang_name, download_method=None, processors='tokenize,pos,lemma', use_gpu=True,  **{'tokenize': {'mode': 'A'}})
        return self._nlp_stanza

    @property
    def corpus_occurrences(self):
        if self.corpus_occurrences_ is None:
            try:
                # TODO: Dependency to parent folder not good
                with open(f'{PROJECT_ROOT}/corpus_resources/{self.source_lang}.pkl', "rb") as f:
                    self.corpus_occurrences_ = pickle.load(f)
            except:
                self.corpus_occurrences_ = Counter()
        return self.corpus_occurrences_

    def collect_file_paths(self, path, excluded_files = None, file_match_pattern = None, dir_match_pattern = None):
        compiled_file_pattern = re.compile(file_match_pattern) if file_match_pattern else None
        compiled_dir_pattern = re.compile(dir_match_pattern) if dir_match_pattern else None
        file_paths = set()
        path = Path(path)
        if path.is_file():
            if not excluded_files or Path(path).stem not in excluded_files and compiled_file_pattern.match(Path(path).name):
                file_paths.add(path)
        elif path.is_dir():
            for root, dirs, files in os.walk(path):
                if compiled_dir_pattern:
                    dirs[:] = [d for d in dirs if compiled_dir_pattern.match(d)]
                for filename in files:
                    if not compiled_file_pattern or compiled_file_pattern.match(Path(filename).name):
                        if not excluded_files or Path(filename).stem not in excluded_files:
                            file_paths.add(Path(root, filename))
        else:
            raise RuntimeError(f'Path does not exist: "{path}"')
        return file_paths

    def detect_languages(self, file_contents):
        """
        Currently used to assert integrity of files (only one language used)

        """
        part_length = len(file_contents) // 4
        parts = [file_contents[i * part_length:(i + 1) * part_length] for i in range(4)]

        detected_languages = set()
        for part in parts:
            try:
                lang = detect(part)
                detected_languages.add(lang)
            except LangDetectException:
                continue
        return detected_languages

    def clean_string(self, string, chars_to_filter):
        string = ''.join(
            c for c in string if unicodedata.category(c) not in {'Cc', 'Cf', 'Cs', 'Co', 'Cn', 'Zl', 'Zp'} or c in {'\n'}
        )
        if chars_to_filter:
            for char_to_filter in chars_to_filter:
                replacement = " "
                if isinstance(char_to_filter, tuple):
                    char_to_filter, replacement = char_to_filter
                string = re.sub(char_to_filter, replacement, string)
        string = pattern.sub('', string)
        return string.strip()

    def open_files(self,
                   file_paths: List[Union[str, Path]],
                   replace_chars: List[str] = None,
                   subdomain_name: str = None,
                   assert_coherence: bool = False,
                   index_separators: Iterable[str] = None):

        """
        Extract raw text from list of files.

        Valid formats:
        - Any raw text (.txt etc.)
        - Subtitles (.ass, .srt)
        - Websites (.html)

        Parameters:
        -----------

        file_paths: list of files to process
        replace_chars: List of characters to be removed when processing
        subdomain_name: Allows for special parcing of content on specific websites in parse_html
            Currently only have rule for ['wikipedia']
        assert_coherence: check if files have sentences in more than one language, skip if that is the case.
        index_separators: Custom delimiter to text into small segments.
            Only relevant if we want to create example sentences from this in the db.
            Subtitles are already segmented in small chunks by default.
            If left empty, stanza sentence segementation will be used.

        """
        files = []
        text_indices = None
        sub_indices = None
        processed_words = {} # storing past results
        for file_path in file_paths:
            with open(file_path, 'rb') as f:
                raw_data = f.read()
            encoding_result = chardet.detect(raw_data)
            encoding = encoding_result['encoding']
            if not encoding:
                encoding = 'utf-8'
            try:
                raw_file_contents = raw_data.decode(encoding, errors='replace')
            except LookupError as e:
                logger.error(f'Decode error: {e}')
                continue
            file_contents = None
            if file_path.suffix in ['.ass', '.srt']:
                if index_separators:
                    raise RuntimeError('index_separators not used when parsing subs')
                try:
                    file_contents, text_indices, sub_indices = self.parse_subs(raw_file_contents, replace_chars=replace_chars)
                except Exception as e:
                    print(f'Error: "{file_path}": {type(e).__name__} - {e}')
            else:
                if file_path.suffix in ['.html']:
                    raw_file_contents = self.parse_html(raw_file_contents, subdomain_name = subdomain_name)
                file_contents = self.clean_string(raw_file_contents, replace_chars)

            if file_contents:
                if assert_coherence:
                    langs = self.detect_languages(file_contents)
                    if len(langs)>1:
                        logger.error(f"Skipping file {file_path}, Multiple languages detected: {langs}")
                        continue
                logger.debug(f'appending file {file_path}')
                logger.debug(f'file_contents1: {len(file_contents)}, {file_contents}')
                logger.debug(f'text_indices: {text_indices}')
                logger.debug(f'sub_indices: {sub_indices}')
                files.append(File(self.source_lang,
                                file_contents,
                                processed_words=processed_words,
                                path=file_path,
                                text_indices=text_indices,
                                sub_indices=sub_indices,
                                index_separators=index_separators))
        return sorted(files, key=lambda x: x.path.name)

    def process_files(self, files: List['File'], get_indices=False, n_threads=None, max_retries=3):
        """
        Concurrently extract lemmas from files

        """
        def process_file(file):
            retry_count = 0
            while retry_count < max_retries:
                try:
                    file.analyze_lemmas(self.nlp_stanza, get_indices=get_indices)
                    break
                except self.torch.cuda.OutOfMemoryError as e:
                    logger.error(f"CUDA out of memory error: {e}. Attempt {retry_count + 1} of {max_retries}.")
                    logger.error("file", file.content)
                    self.torch.cuda.empty_cache()
                    break

        with ThreadPoolExecutor(max_workers=n_threads if n_threads else 5) as executor:
            futures = [executor.submit(process_file, file) for file in files]
            progress = tqdm(as_completed(futures), total=len(futures), desc="Extracting lemmas")
            for future in progress:
                future.result()

    def parse_html(self, file_contents: str, subdomain_name = None) -> str:
        """
        Clean html files to extract text (relevant for scraped pages)

        """
        soup = BeautifulSoup(file_contents, 'lxml')
        if subdomain_name == 'wikipedia':
            if soup.find('meta', attrs={'http-equiv': 'Refresh'}):
                return ""
            body_content_div = soup.find('div', id='bodyContent')
            text = ""
            if body_content_div:
                paragraphs = body_content_div.find_all('p')
                for paragraph in paragraphs:
                    text += paragraph.get_text() + "\n"
            clean_text = re.sub(r'\[.*?\]', '', text)
            clean_text = re.sub(r'[^\x00-\x7F]+', '', clean_text)
        else:
            clean_text = soup.get_text()
        return clean_text.strip()

    def parse_subs(self, file_contents: str, replace_chars: List[str] = None) -> str:
        """
        Extract and clean text from sub files

        """
        logger.debug(f'parse_subs: {file_contents}')
        try:
            subs = pysubs2.SSAFile.from_string(file_contents)
        except Exception as e:
            subs = pysubs2.SSAFile.from_string(file_contents, format="srt")
        text_list = []

        # TODO: Currently manually assembled from trial and error with different files, is there an automated way to solve this?
        filter_expressions = {
            'jp': [r'《', r'》',r'（.*?）', r'\[.*\]', r'\{.*?\}', r'\(.*?\)', '\u3000', '\n', '“','”','「', '」', r'\\N', r'\n', r'\n' '…', '→','〉', '〈', '‥', '⁉', r'[\u2460-\u2473]', '‪', '‬', r'\r', r'  '],
            'de': [r'\{.*?\}', '\.', ',', '\!', ';', '\"', "\'", '\?', ('\n', ' '), ':', r'\\N', '%']
        }
        def is_jp_dialogue(text):
            non_dialogue_patterns = [r'---', r'翻译:', r'www.', '字幕組', '翻译', '校对', '后期', '广告', '制作成员']
            if any(re.search(pattern, text) for pattern in non_dialogue_patterns):
                return False
            if re.search(r'[\u3000\u3040-\u30ff\u4e00-\u9fff]+', text):
                return True
            return True
        text_indices = []
        sub_indices = []
        current_index = 0
        for line in subs:
            logger.debug(f'Cleaning line {[line.text]}')
            text = self.clean_string(line.text, filter_expressions.get(self.source_lang, []) + (replace_chars if replace_chars else []))
            logger.debug(f'Cleaned  line: {[text]}')
            if (not self.source_lang=='jp' or is_jp_dialogue(text)) and text:
                start_time = round(line.start / 1000)
                end_time = round(line.end / 1000)
                start_index = current_index
                text_list.append(text)
                end_index = start_index+len(text)
                current_index += len(text) + 1
                text_indices.append((start_index, end_index))
                sub_indices.append((start_time, end_time))
        final_text = ' '.join(text_list).strip()
        logger.debug(f'text_indices: {text_indices}')
        logger.debug(f'sub_indices: {sub_indices}')
        return final_text, text_indices, sub_indices

class File:
    def __init__(self,
                 source_lang,
                 content: List[str],
                 processed_words: Dict[str, str] = None,
                 path = None,
                 text_indices: List[Tuple[int, int]] = None,
                 sub_indices: List[Tuple[int, int]] = None,
                 index_separators: str = None):
        self.source_lang = source_lang
        self.path = path
        self.content = content
        self.processed_words = processed_words if processed_words else {}
        self.text_indices = text_indices
        self.sub_indices = sub_indices # text indices -> time on screen
        self.index_separators = index_separators

        # Variables to be written to when we are adding this source to the main db (TODO: move this from ankipan to ankipan_db)
        self.text_segment_components = OrderedDict() #This has the same keys as sub_indices, but points to a list of (lemma, word)
        self.word_count = Counter()
        self.stanza_segments_by_indices = None

    @property
    def nlp_hanta(self):
        global _nlp_hanta
        if not _nlp_hanta:
            from HanTa import HanoverTagger as ht
            _nlp_hanta = ht.HanoverTagger(hanta_map[self.source_lang]) if self.source_lang in hanta_map.keys() else None
        return _nlp_hanta

    def parse_stanza_segments_by_indices(self, nlp_stanza):
        """
        In-place parsing of self.content into stanza_segments_by_indices

        """

        if self.text_indices and self.index_separators:
            raise RuntimeError(f'index_separators "{self.index_separators}" would override text_indices {self.text_indices}')
        try:
            doc = nlp_stanza(self.content)
            stanza_sentences = doc.sentences
        except Exception as e:
            if len(self.content) > 10000:
                self.content = self.content[:10000]
                print("content", self.content)
            raise e
        if not self.text_indices and self.index_separators:
            indices = []
            separators = '|'.join(map(re.escape, self.index_separators))
            pattern = f"({separators})"
            for sentence in stanza_sentences:
                sentence_start = sentence.tokens[0].start_char
                sentence_end = sentence.tokens[-1].end_char
                sentence_text = self.content[sentence_start:sentence_end]
                segments = re.split(pattern, sentence_text)
                start = sentence_start

                for i in range(len(segments)):
                    trimmed_segment = segments[i].strip()
                    if trimmed_segment:
                        segment_start = self.content.find(trimmed_segment, start)
                        segment_end = segment_start + len(trimmed_segment)
                        indices.append([segment_start, segment_end])
                        start = segment_end
            if indices:
                indices = sorted(indices)
                merged_indices = []
                current_start, current_end = indices[0]
                for start, end in indices[1:]:
                    if start <= current_end:
                        current_end = max(current_end, end)
                    else:
                        merged_indices.append([current_start, current_end])
                        current_start, current_end = start, end
                merged_indices.append([current_start, current_end])
                self.text_indices = merged_indices
        if self.text_indices:
            current_index = 0
            index_tree = IntervalTree(Interval(start, end, []) for start, end in self.text_indices)
            for sentence in stanza_sentences:
                for word in sentence.words:
                    if word.start_char: current_index = word.start_char
                    overlapping_intervals = index_tree[current_index]
                    if overlapping_intervals:
                        first_interval = next(iter(overlapping_intervals))
                        first_interval.data.append(word)
            sorted_intervals = sorted(index_tree, key=lambda interval: interval.begin)
            stanza_segments = OrderedDict(((interval.begin, interval.end), interval.data) for interval in sorted_intervals)
        else:
            stanza_segments = {
                (sentence.tokens[0].start_char, sentence.tokens[-1].end_char):
                [word for word in sentence.words if word.lemma] for sentence in stanza_sentences
            }
        return stanza_segments

    def analyze_lemmas(self, nlp_stanza, get_indices = False):
        """
        In-place parsing of lemmas in self.stanza_segments_by_indices into self.text_segment_components

        """

        self.stanza_segments_by_indices = self.parse_stanza_segments_by_indices(nlp_stanza)
        for text_char_indices, stanza_segment in self.stanza_segments_by_indices.items():
            metadata = []
            # TODO: collect words in list first, then fetch available lemmas from db
            for i, stanza_token in enumerate(stanza_segment):
                word = stanza_token.text
                if word in self.processed_words:
                    self.word_count[self.processed_words[word]] = self.word_count.get(self.processed_words[word], 0) + 1
                    if get_indices:
                        metadata.append({
                            'word': word,
                            'lemma': self.processed_words[word]})
                else:
                    lemma = None

                    if self.source_lang=='jp':
                        if stanza_token.lemma and re.search(r'[\u4E00-\u9FD0]+', stanza_token.lemma) and re.search(r'[\u4E00-\u9FD0]+', stanza_token.text):
                            lemma = stanza_token.lemma
                    elif stanza_token.pos in ["NOUN", "VERB", "ADJ", "ADV", "AUX", "PROPN"] and stanza_token.xpos != 'NE' and stanza_token.lemma and stanza_token.lemma.isalpha():
                        if self.source_lang=='de':
                            try:
                                hanta_lemmas = self.nlp_hanta.tag_sent([token.text for token in stanza_segment])
                                hanta_lemma = hanta_lemmas[i] if i < len(hanta_lemmas)-1 and hanta_lemmas[i][0] == stanza_token.text else \
                                            self.nlp_hanta.tag_sent([stanza_token.text])[0]
                                lemma = hanta_lemma[1] if len(hanta_lemma)>=3 else stanza_token.lemma
                            except IndexError as e:
                                print(f"Hanta Error: {e}")
                                lemma = stanza_token.lemma
                        else:
                            lemma = stanza_token.lemma
                    if lemma:
                        self.word_count[lemma] = self.word_count.get(lemma, 0) + 1
                        if get_indices:
                            metadata.append({
                                'word': word,
                                'lemma': lemma,
                                })
                        if stanza_token.text not in self.processed_words:
                            self.processed_words[stanza_token.text] = lemma
                # print("====================================================")
                # print("stanza_token.lemma",stanza_token.lemma)
                # print("stanza_token.text",stanza_token.text)
                # print("stanza_token.pos",stanza_token.pos)
                # print("stanza_token.xpos",stanza_token.xpos)
                # print("hanta_lemma",hanta_lemma)
            if get_indices:
                self.text_segment_components[text_char_indices] = metadata



# Disclaimer: The current scraping methods are only implemented for sources where scraping is not explicitly prohibited.
# The downloaded data is used strictly for educational purposes, and is only stored locally for each individual user who is creating their own flashcards.
# If you are the owner of any of those sources and would like your scraping code removed, please message me and I will do so right away. (daniel.mentock@gmail.com)

import requests
from bs4 import BeautifulSoup
import json
from collections import OrderedDict
from pathlib import Path
from jinja2 import Environment, FileSystemLoader, select_autoescape
import html
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm.notebook import tqdm
import time
import random
import re
import fnmatch
import pykakasi
import markdown
import logging

from typing import List, Dict, Union, Iterable, get_args

from ankipan import HTML_TEMPLATE_DIR, ANKIPAN_DB_ADDR

logger = logging.getLogger(__name__)

deeptrans_lang_map = {
    'jp': 'japanese',
    'en': 'english',
    'fr': 'french',
    'de': 'german'
}

google_dict_lang_map = {
    'jp': 'ja',
}

wikitionary_lang_map = {
    'jp': 'ja',
}

ContentNode = Union['HeaderAndContent','Collapsible', 'DictEntries', 'RawHtml']
EnumNode = Union['Enumeration', 'BulletEnumeration', 'BracketedList', 'ConcatenatedSections', 'SpanEnumeration']
ScalarNode = Union['Sentence','Placeholder', 'YoutubeEmbed', 'Wikitionary']
AnyNode = Union[EnumNode, ContentNode, ScalarNode, str]

env = Environment(
    loader=FileSystemLoader(HTML_TEMPLATE_DIR),
    autoescape=select_autoescape(['html']),
)
env.filters['render'] = lambda value: str(value)

# color codes: https://htmlcolorcodes.com/color-names/
class CardSection:
    def __init__(self,
                 display_name: str,
                 color: str,
                 content: AnyNode,
                 url: str = None,
                 is_open = False):
        self.display_name = display_name
        self.color = color
        self.content = content
        self.url = url
        self.is_open = is_open

    def as_dict(self):
        def generate_dict(field_element):
            class_name = field_element.__class__.__name__
            if items:= field_element.items() if isinstance(field_element, dict) else \
                       field_element.__dict__.items() if hasattr(field_element, '__dict__') else None:
                result = OrderedDict()
                for attr, value in items:
                    result[attr] = generate_dict(value)
            elif isinstance(field_element, list) or isinstance(field_element, set):
                result = [generate_dict(item) for item in field_element]
            else:
                result = field_element
            return (class_name, result)
        return generate_dict(self)

    @staticmethod
    def from_dict(data, is_type_description_node = True):
        if is_type_description_node:
            if data[0] in [CardSection.__name__] + \
                          [ref.__forward_arg__ for ref in get_args(EnumNode)] + \
                          [ref.__forward_arg__ for ref in get_args(ContentNode)]+ \
                          [ref.__forward_arg__ for ref in get_args(ScalarNode)]:
                ClassRef = globals()[data[0]]
                return ClassRef(**CardSection.from_dict(data[1], is_type_description_node = False))
            else:
                return CardSection.from_dict(data[1], is_type_description_node = False)
        elif isinstance(data, dict):
            return OrderedDict({key: CardSection.from_dict(val) for key, val in data.items()})
        elif isinstance(data, list):
            return [CardSection.from_dict(item) for item in data]
        else:
            return data

    def __str__(self):
        if str(self.content):
            template = env.get_template('flashcard_field.html')
            return template.render(flashcard_field=self)
        else:
            return ''

    def __bool__(self):
        return bool(self.content)

class HeaderAndContent:
    def __init__(self, header: str,
                 content: AnyNode,
                 header_level: int = None,
                 header_color: str = 'white'):
        self.header = header
        self.content = content
        self.header_level = header_level if header_level else 4
        self.header_color = header_color

    def __str__(self):
        template = env.get_template('header_and_content.html')
        return template.render(header_and_content=self)

    def __bool__(self):
        return bool(self.content)

class Collapsible(HeaderAndContent):
    def __init__(self, header: str,
                 content: AnyNode,
                 header_level: int = None,
                 is_open: bool = False,
                 header_color: str = 'white',
                 summary_size: int = None,
                 light: bool = False):
        super().__init__(header, content, header_level, header_color)
        self.is_open = is_open
        self.summary_size = summary_size
        self.light = light

    def __str__(self):
        template = env.get_template('collapsible.html')
        return template.render(collapsible=self)

    def __bool__(self):
        return bool(self.content)

class DictEntries:
    def __init__(self,
                 definitions: Dict[str, AnyNode]):
        self.definitions = definitions

    def __str__(self):
        template = env.get_template('dict_entries.html')
        return template.render(dict_entries=self)

    def __bool__(self):
        return bool(self.definitions)

class Enumeration:
    def __init__(self, entries: Iterable[AnyNode]):
        self.entries = entries

    def __str__(self):
        return '\n'.join([f'{str(item)}' for item in self.entries if item])

    def __bool__(self):
        return bool(self.entries)

class SpanEnumeration(Enumeration):
    def __str__(self):
        return '\n'.join([f'<span style="padding: 0; margin: 0;"> {str(item)}</span><br>' for item in self.entries if item])

class BracketedList(Enumeration):
    def __str__(self):
        if self.entries:
            return '(' + ', '.join([f'{str(item)}' for item in self.entries if item]) + ')'
        else:
            return ''

class ConcatenatedSections(Enumeration):
    def __str__(self):
        return ' '.join([f'{str(item)}' for item in self.entries if item])

class CommaSeparatedSections(Enumeration):
    def __str__(self):
        return '; '.join([f'{str(item)}' for item in self.entries if item])

class BulletEnumeration(Enumeration):
    def __str__(self):
        template = env.get_template('bullet_enumeration.html')
        return template.render(bullet_enumeration=self)

class RawHtml():
    def __init__(self, html):
        self.html = html

    def __str__(self):
        return f'<div style="text-align: center !important;">\n{self.html}\n</div>'

    def __bool__(self):
        return True if self.html else False

class Sentence():
    def __init__(self, sentence, word, youtube_hash = None):
        self.sentence = sentence
        self.word = word

    def __str__(self):
        return re.sub(
            rf'(\s*){re.escape(self.word.strip())}(\s*)',
            lambda m: f'{m.group(1)}&#8239;<div style="color:red !important; display:inline;">{html.escape(self.word.strip())}</div>&#8239;{m.group(2)}',
            html.escape(str(self.sentence)))

    def __bool__(self):
        return True if self.sentence and self.word else False

class YoutubeEmbed():
    def __init__(self, youtube_hash, start_s, end_s):
        self.youtube_hash = youtube_hash
        self.start_s = start_s
        self.end_s = end_s

    def __str__(self):
        template = env.get_template('youtube_embed.html')
        return template.render(youtube_embed=self)

    def __bool__(self):
        return True if self.youtube_hash else False

class Wikitionary():
    def __init__(self, content):
        self.content = content

    def __str__(self):
        template = env.get_template('wikitionary.html')
        return template.render(wikitionary=self)

    def __bool__(self):
        return True if self.content else False

class Scraper:
    def __init__(self, source_lang, native_lang):
        self.source_lang = source_lang
        self.native_lang = native_lang
        self.timeout = 25
        self.unsplash_access_key = "6EwImfEiXvVj0QxjzVV16R3hY13ar20I5HPRWg79kgU"
        self.kks = pykakasi.kakasi()

    def download_flashcard_fields(self,
                                  missing_fields: Dict[str, List[str]],
                                  known_sources_by_subdomain: Dict[str, List[str]] = None,
                                  save_func=None):
        """
        Called by Collection.fetch to trigger download of contents for this word

        TODO: print available fields

        missing_fields: dict, field_name -> list of lemmas
        scraper: Scraper module passed from collection
            Only one instance is required instead of one for each card
            Especially since it only contains static methods

        """
        logger.debug(f'download_flashcard_fields: {missing_fields.keys()}')
        downloaded_fields = {}
        if has_sentences:= any([field_name for field_name in missing_fields.keys() if field_name.startswith('sentences_')]):
            task_id = self.trigger_sentences(
                {'_'.join(field_name.split('_')[1:]): words for field_name, words in missing_fields.items()
                    if field_name.startswith("sentences_")}, known_sources_by_subdomain = known_sources_by_subdomain)

        tasks = []
        for field_name, words in {field_name: words for field_name, words in missing_fields.items()
                                    if not field_name.startswith("sentences_")}.items():
            try:
                func = getattr(self, f'fetch_{field_name}')
            except AttributeError as e:
                raise RuntimeError(f'Section name "{field_name}" not specified in scraper.py module.')

            for word in words:
                tasks.append((field_name, word, func))
        random.shuffle(tasks)
        word_to_future = {}
        downloaded_fields = {}
        with ThreadPoolExecutor(max_workers=5) as executor:
            for field_name, word, func in tasks:
                future = executor.submit(func, word)
                word_to_future[(field_name, word)] = future

            for i, ((field_name, word), future) in enumerate(word_to_future.items()):
                for attempt in range(5):
                    try:
                        result = future.result()
                        continue
                    except Exception as e:
                        logger.error(f'Fail to collect field {field_name} for word "{word}": {e}. Waiting before retry...')
                        time.sleep(150)
                if field_name not in downloaded_fields:
                    downloaded_fields[field_name] = {}
                downloaded_fields[field_name][word] = result
                print("finished", field_name, word)
                if not i+1%5 and save_func:
                    print(f"saving ({save_func})")
                    save_func()

        if has_sentences:
            sentences_by_subdomains = self.download_sentences(task_id)
            for subdomain, sentences in sentences_by_subdomains.items():
                downloaded_fields[f'sentences_{subdomain}'] = sentences
        return downloaded_fields

    def trigger_sentences(self,
                          lemmas_by_subdomain: Dict[str, List[str]],
                          known_sources_by_subdomain: Dict[str, List[str]]) -> List[int]:
        available_sources = self.get_available_sources()
        unknown_sources = []
        for subdomain, source_names in known_sources_by_subdomain.items():
            for source_name in source_names:
                if source_name not in available_sources.get(subdomain, {}):
                    unknown_sources.append((subdomain, source_name))
        if unknown_sources:
            raise RuntimeError(f'Invalid known sources in collection: "{unknown_sources}"')
        chunk_size = 5
        task_ids = []
        lemma_chunks = [list(lemmas_by_subdomain.items())[i:i + chunk_size] for i in range(0, len(lemmas_by_subdomain), chunk_size)]
        for lemma_chunk in lemma_chunks:
            data = {'lemmas_by_subdomain': dict(lemma_chunk), 'known_sources_by_subdomain': known_sources_by_subdomain, 'source_lang': self.source_lang, 'native_lang': self.native_lang}
            response = requests.post(f'{ANKIPAN_DB_ADDR}/api/sentences', json=data).json()
            task_ids.append(response['task_id'])
        return task_ids

    def download_sentences(self, task_ids: List[int]):
        def generate_sentences_overview(sentence_overview_by_lemmas):
            results = {}
            for lemma, sentence_overview_fields in sentence_overview_by_lemmas.items():
                fields = []
                for field, overview in sentence_overview_fields.items():
                    fields.append(Collapsible(field, RawHtml(markdown.markdown(overview))))
                results[lemma] =  CardSection(f'Overview', 'Crimson', SpanEnumeration(fields), is_open = True)
            return results
        i = 0
        sentences_by_subdomain = {}
        while task_ids:
            for idx in range(len(task_ids) - 1, -1, -1):
                task_id = task_ids[idx]
                status_url = f'{ANKIPAN_DB_ADDR}/api/sentences/status/{task_id}'
                try:
                    response = requests.get(status_url).json()
                except requests.RequestException as e:
                    raise RuntimeError(f"HTTP request failed: {e}")
                if response['status'] == 'SUCCESS':
                    task_ids.pop(idx)
                    examples_results = {k: v for k, v in response['result'].items() if k!='_overview'}
                    sentences_by_subdomain.update(self.generate_sentences(examples_results))
                    if (sentence_overview_by_lemmas:=response['result'].get('_overview')):
                        sentences_overview = {}
                        for lemma, sentence_overview_fields in sentence_overview_by_lemmas.items():
                            fields = []
                            for field, overview in sentence_overview_fields.items():
                                fields.append(Collapsible(field, RawHtml(markdown.markdown(overview))))
                            sentences_overview[lemma] =  CardSection(f'Overview', 'Crimson', SpanEnumeration(fields), is_open = True)
                        sentences_by_subdomain.setdefault('_overview', {}).update(sentences_overview)

                elif response['status'] == 'FAILURE':
                    raise RuntimeError(f"Task failed: {response['result']}")
            i += 1
            print(f"Waiting for db sentences query to finish, iteration {i}")
            time.sleep(5)
        logger.debug(f'finish download_sentences')
        return sentences_by_subdomain

    def generate_sentences(self, db_result):
        logger.debug(f'generate_sentences: {db_result}')
        sentences_by_subdomain = {}
        for subdomain, sentences_by_lemmas in db_result.items():
            words_and_sentences = {}
            for lemma, known_and_unknown_sentences_by_source in sentences_by_lemmas.items():
                def process_sentences(entries_and_sentences_by_source: Dict[str, Dict[str, List[str]]], subdomain, color='white'):
                    if not entries_and_sentences_by_source: return []
                    rendered_sentences = []
                    for source_name, sentences_by_source_entry in entries_and_sentences_by_source.items():
                        source_sentences = []
                        for source_entry_name, source_entry_sentences in sentences_by_source_entry.items():
                            entry_sentences = []
                            for i, source_entry_sentence in enumerate(source_entry_sentences):
                                sentence_components = []
                                word = source_entry_sentence['word']
                                entries = source_entry_sentence['entries']
                                entries_field = []
                                if subdomain == 'youtube':
                                    try:
                                        start_s = source_entry_sentence['start_s']
                                        end_s = source_entry_sentence['end_s']
                                        entry_title = ' '.join(source_entry_name.split('_')[1:]).replace('-', ' ')
                                        youtube_hash = source_entry_name.split('_')[0]
                                        entries_field.append(YoutubeEmbed(youtube_hash, start_s, end_s))
                                    except:
                                        pass
                                else:
                                    entry_title = source_entry_name.replace('_', ' ').replace('-', ' ')
                                for j, entry in enumerate(entries):
                                    translation = '<To be translated>' if source_entry_sentence['translations'][j] == None else source_entry_sentence['translations'][j]
                                    if self.source_lang=='jp':
                                        translation+='<br>'+self.add_furigana(entry)
                                    logger.debug(f'entry: {entry}')
                                    logger.debug(f'translation: {translation}')
                                    entries_field.append(Collapsible(
                                        Sentence(entry, word), translation, summary_size='14px', light=True))
                                    logger.debug(f'entries_field: {entries_field}')

                                sentence_components.append(Enumeration(entries_field))
                                entry_sentences.append(Collapsible(entry_title, Enumeration(sentence_components), header_level=6, is_open=True, header_color = 'Silver'))
                            source_sentences.append(Enumeration(entry_sentences))
                        rendered_sentences.append(Collapsible(source_name, Enumeration(source_sentences), header_color=color))
                    return rendered_sentences
                parsed_sentences = []
                if known_sources := known_and_unknown_sentences_by_source.get('entries_from_known_sources'):
                    known_sentences = process_sentences(known_sources, subdomain, color = '31EC81')
                    logger.debug(f"known_sources: {known_sources}, {len(known_sentences)}")
                    parsed_sentences.extend(known_sentences)
                if unknown_sources := known_and_unknown_sentences_by_source.get('entries_from_unknown_sources'):
                    extra_sentences = process_sentences(unknown_sources, subdomain)
                    logger.debug(f"unknown_sources: {unknown_sources}, {len(extra_sentences)}")
                    parsed_sentences.extend(extra_sentences)
                words_and_sentences[lemma] = CardSection(subdomain.title(), 'DarkOrange', Enumeration(parsed_sentences),
                                                        is_open = True)
            sentences_by_subdomain[subdomain] = words_and_sentences
        logger.debug(f'finish generate_sentences')
        return sentences_by_subdomain

    def fetch_jisho(self, word: str) -> Dict[str, Dict[str, str]]:
        url="https://jisho.org/search/{}".format(word)
        headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'}
        vstr=requests.get(url, headers=headers, timeout=self.timeout).content
        soup = BeautifulSoup(vstr,features="html.parser",from_encoding='utf8')
        rows = soup.findAll('div', {"class":"concept_light clearfix"})
        jisho = OrderedDict()
        for row in rows:
            furigana = [span.get_text() for span in row.find_all('span', class_='kanji')]
            text_jp_ = row.findAll('span', {"class":"text"})
            text_jp = str(text_jp_).replace('<span class="text">',"").replace("<span>","").replace("</span>","").replace(" ","").replace("[","").replace("]","").replace("\n","").strip()
            text_target_ = row.findAll('span', {"class":"meaning-meaning"})
            text_target = str(text_target_).replace('<span class="meaning-meaning">',"").replace("<span>","").replace("</span>","").replace("[","").replace("]","").replace("\n","").strip()
            if "<span" in text_target:
                text_target = text_target.split("<span")[0]
            if "\n" in text_target:
                text_target = text_target.split("\n")[0]
            if text_jp not in jisho.keys() and furigana and text_jp and text_target:
                jisho[text_jp] = ConcatenatedSections([text_target, BracketedList(furigana)])
        return CardSection('Jisho.org', 'Violet', DictEntries(jisho), url = url)

    def fetch_tatoeba(self, word: str) -> Dict[str, List[str]]:
        lang_mapping = {
            'jp': 'jpn',
            'en': 'eng',
            'de': 'deu',
            'fr': 'fra'
        }
        def fetch(source_lang, native_lang):
            max_pages = 3
            tatoeba=OrderedDict()
            url= f"https://tatoeba.org/eng/sentences/search?from={lang_mapping[self.source_lang]}&query={word}&to={lang_mapping[self.native_lang]}"
            vstr=requests.get(url, timeout=self.timeout).content
            soup = BeautifulSoup(vstr,features="html.parser",from_encoding='utf8')
            paging = soup.find('ul', class_='paging')
            if paging:
                li_tags = paging.find_all('li', {'class': lambda x: x != 'next' and x != 'ellipsis'})
                # Get the text of the last li tag (which will be the last page number)
                last_page = min(int(li_tags[-1].a.get_text()), max_pages)
                # last_page = 2
            else:
                last_page = 1
            for page in range(1, last_page+1):
                if page>1:
                    url= f"https://tatoeba.org/eng/sentences/search?from={lang_mapping[self.source_lang]}&query={word}&to={lang_mapping[self.native_lang]}&page={page}"
                    vstr=requests.get(url, timeout=self.timeout).content
                    soup = BeautifulSoup(vstr,features="html.parser",from_encoding='utf8')
                rows = soup.findAll('div', {"class":"sentence-and-translations"})
                for row in rows:
                    if row:
                        ng_init = row['ng-init']
                        comma_pos = ng_init.find(',')
                        end_bracket_pos = ng_init.rfind(']')
                        json_str = f'[{ng_init[comma_pos+1:end_bracket_pos].strip()}]]'
                        data = json.loads(json_str)
                        example = data[0]['text']
                        translations = []
                        for translation in data[0]['translations']:
                            if translation:
                                translations.append(translation[0]['text'])
                        tatoeba[example] = Enumeration(translations)
            return tatoeba
        examples = fetch(self.source_lang, self.native_lang) if self.native_lang!='en' else OrderedDict()
        examples.update(fetch(self.source_lang, 'en'))


        return CardSection('Tatoeba.org', 'Tomato', DictEntries(examples),
            url = f"https://tatoeba.org/eng/sentences/search?from={lang_mapping[self.source_lang]}&query={word}&to={lang_mapping[self.native_lang]}")

    def fetch_urban(self, word: str) -> List[str]:
        logger.debug(f'fetch_urban: {word}')

        url = f"http://api.urbandictionary.com/v0/define?term={word}"
        response = requests.get(url)
        dictionary = json.loads(response.text)['list']
        definitions = {}
        for definition in dictionary:
            definitions[definition['word']] = {'thumbs': definition['thumbs_up'], 'definition': definition['definition']}

        return CardSection(
            'Urban Dictionary',
            'Violet',
            DictEntries(
                OrderedDict({
                    word: info["definition"] for word, info in sorted(definitions.items(), key=lambda item: item[1]['thumbs'], reverse=True)
            })),
            url = url)

    def fetch_sprachnudel(self, word: str) -> Dict[str, Union[List[str], str]]:
        logger.debug(f'fetch_sprachnudel: {word}')

        url = f'https://www.sprachnudel.de/woerterbuch/{word}'
        headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'}
        html_content = requests.get(url, headers=headers, timeout=self.timeout).content
        if 'Hast du dich verirrt?' in str(html_content):
            content = ''
        else:
            soup = BeautifulSoup(html_content, 'html.parser', from_encoding='utf8')
            main_meaning = secondary_meaning = examples = ''
            for div in soup.find_all('div'):
                text = div.get_text()
                if 'Hauptbedeutung' in text:
                    main_meaning = text.split('Hauptbedeutung')[1].split('Nebenbedeutung')[0].strip()
                elif 'Bedeutung (Definition)' in text and not main_meaning:
                    main_meaning = text.split('Bedeutung (Definition)')[-1].split('Deine Bedeutung')[0].strip()
                if 'Nebenbedeutung' in text:
                    secondary_meaning = text.split('Nebenbedeutung')[1].split('Assoziative Bedeutungen')[0].strip()
                if 'Beispielsätze' in text:
                    examples = [example.strip() for example in text.split('Beispielsätze')[1].split('Dein Beispielsatz')[0].strip().split('\n') if example.strip()]

            def extract_associative_meanings(soup):
                associative_headings = soup.find_all('h3', text='Assoziative Bedeutungen')
                for heading in associative_headings:
                    next_ul = heading.find_next_sibling('ul')
                    if next_ul:
                        associative_meanings = [li.get_text(strip=True) for li in next_ul.find_all('li')]
                        return associative_meanings
                return []
            content = Enumeration([
                HeaderAndContent('Main Meaning', main_meaning),
                HeaderAndContent('Secondary Meaning', secondary_meaning),
                HeaderAndContent('Examples', BulletEnumeration(examples)),
                HeaderAndContent('Associative Meanings', BulletEnumeration(extract_associative_meanings(soup)))
            ])
        return CardSection('Sprachnudel.de', 'Bisque', content, url = url)

    def fetch_wadoku(self, word):
        url = f'https://www.wadoku.de/search/{word}'
        headers = {'User-Agent': 'Mozilla/5.0'}
        html_content = requests.get(url, headers=headers).content
        soup = BeautifulSoup(html_content, 'html.parser')
        table = soup.find('table', id='resulttable')
        dict_entries = {}
        if table:
            rows = table.find_all('tr')
            for row in rows:
                cells = row.find_all('td')
                if len(cells) >= 3:
                    japanese_parts = cells[1].find_all('span', class_='orth')
                    japanese_word = ''.join(part.get_text(strip=True) for part in japanese_parts)
                    furigana_parts = cells[1].find_all('span', class_='reading')
                    furigana = ''.join(part.get_text(strip=True) for part in furigana_parts)
                    japanese_formatted = f'{japanese_word} ({furigana})' if furigana else japanese_word
                    german_translation = ''
                    german_translation_field = cells[2].find('div', class_='d')
                    if german_translation_field:
                        german_translation = RawHtml(str(german_translation_field))
                    else:
                        german_translation = RawHtml(str(cells[2]))
                    dict_entries[japanese_formatted] = german_translation
        return CardSection('Wadoku.de', 'Lavender', DictEntries(dict_entries), url=url)

    def _fetch_wikitionary(self, lang, word):
        logger.debug(f'fetch_wikitionary: {word}')

        url = f'https://{lang}.m.wiktionary.org/wiki/{word}'
        content = []
        headers = {'User-Agent': 'Mozilla/5.0'}
        html_content = requests.get(url, headers=headers).content
        soup = BeautifulSoup(html_content, 'html.parser')
        if ("Wiktionary does not yet have an entry for mektoub." in soup.text and lang == 'en') or \
           ("ウィクショナリーには現在この名前の項目はありません" in soup.text and lang == 'ja') or \
           ("ne possède pas de page dédiée à cette suite de lettres." in soup.text and lang == 'fr') or \
           ("Dieser Eintrag existiert noch nicht!" in soup.text and lang == 'de'):
            content = ''
        else:
            rows = soup.findAll('div', {"id":"mw-content-text"})
            content = Wikitionary(RawHtml('\n'.join([str(row) for row in rows]).replace('href="/wiki', 'href="https://en.wiktionary.org/wiki')))
        return CardSection(f'{lang}.Wikitionary.org', 'Gray', content, url = url)

    def fetch_wikitionary_en(self, word):
        return self._fetch_wikitionary('en', word)

    def fetch_wikitionary_jp(self, word):
        return self._fetch_wikitionary('ja', word)

    def fetch_wikitionary_fr(self, word):
        return self._fetch_wikitionary('fr', word)

    def fetch_wikitionary_de(self, word):
        return self._fetch_wikitionary('de', word)

    def add_furigana(self, sentence):
        result = []
        for word in self.kks.convert(sentence):
            if word['orig'] != word['hira']:
                result.append(f"{word['orig']} ({word['hira']})")
            else:
                result.append(word['orig'])
        return ''.join(result)

    def translate_google(self, sentences):
        from deep_translator import GoogleTranslator
        translator = GoogleTranslator(source=google_dict_lang_map.get(self.source_lang, self.source_lang),
                                    target=google_dict_lang_map.get(self.native_lang, self.native_lang))
        translations = translator.translate_batch(sentences)
        return translations

    def translate_mymemory(self, sentences):
        from deep_translator import MyMemoryTranslator
        translator = MyMemoryTranslator(source=deeptrans_lang_map[self.source_lang],
                                        target=deeptrans_lang_map[self.native_lang])
        return translator.translate_batch(sentences)

    def translate_pons(self, word) -> List[str]:
        from deep_translator import PonsTranslator
        translator = PonsTranslator(source=deeptrans_lang_map[self.source_lang],
                                        target=deeptrans_lang_map[self.native_lang])
        return translator.translate(word, return_all=True)

    def translate_linguee(self, word) -> List[str]:
        from deep_translator import LingueeTranslator
        translator = LingueeTranslator(source=deeptrans_lang_map[self.source_lang],
                                        target=deeptrans_lang_map[self.native_lang])
        return translator.translate(word, return_all=True)

    def get_available_sources(self, domain_names: List[str] = None) -> Dict[str, Dict[str, List[str]]]:
        data = {'domain_names': domain_names, 'lang': self.source_lang}
        return requests.post(f'{ANKIPAN_DB_ADDR}/api/available_sources', json=data).json()

    # TODO
    # def _fetch_unsplash(self, english_synonyms, count: int = 10) -> List[str]:
    #     unsplash_image_dir = self.project_dir / 'unsplash'
    #     existing_image_ids = [image.split('.')[0] for image in os.listdir(unsplash_image_dir)]
    #     image_ids = []
    #     if not unsplash_image_dir.exists():
    #         unsplash_image_dir.mkdir(parents=True)
    #     url = "https://api.unsplash.com/search/photos"
    #     #todo: print url of all fetched images/number of opened links etc
    #     for word in english_synonyms:
    #         query = {
    #             "query": word,
    #             "client_id": self.unsplash_access_key,
    #             "per_page": count
    #         }
    #         response = requests.get(url, params=query)
    #         if response.status_code == 200:
    #             results = response.json()['results']
    #             for result in results:
    #                 image_url = result['urls']['regular']
    #                 image_id = f'{word}_{result["id"]}'  # Get the image ID from Unsplash
    #                 if image_id not in existing_image_ids:
    #                     print(f'Downloading image {image_id}')
    #                     image_data = requests.get(image_url).content
    #                     image_path = unsplash_image_dir / f"{image_id}.jpg"
    #                     with open(image_path, 'wb') as file:
    #                         file.write(image_data)
    #                 else:
    #                     print(f'Image {image_id} already dowloaded')
    #                 image_ids.append(image_id)
    #             print(f'Downloaded {len(results)} images for "{word}"')
    #         else:
    #             print(f'Error: {response.status_code} when downloading images to {image_path}')
    #     return image_ids



import functools
from jinja2 import Template

from ankipan.scraper import CardSection
from ankipan import HTML_TEMPLATE_DIR, load_json

from typing import Dict, OrderedDict

class Card(OrderedDict):
    def __init__(self,
                 word: str,
                 json_path = None,
                 *,
                 index: int = None,
                 anki_id: int = None,
                 source_references: Dict[str, int] = None):
        """
        Flashcard object.
        Exposes content by inheriting from OrderedDict.
            Key is name of the flashcard field. (e.g. wikitionary or sentences_anime)
            Value is actual content, a CardSection object which contains the downloaded information.
            Values are lazy-loaded from json to prevent a collection with many cards from cluttering memory.

        Parameters
        ----------
        word: Word of the flashcard
        json_path: path to local json path with all fields (saves compute for many cards)
        sources: added sources where this word occurs
        anki_id: optional, filled when card is synced with anki
        source_references: optional, maps source name to number of references
        """
        self.word = word
        self._source_references = source_references if source_references else OrderedDict()
        self._anki_id = anki_id
        self.index = index if index else None
        self.was_modified = False
        self.is_initialized = False
        self.json_path = json_path
        super().__init__({})

    def ensure_initialized(method):
        """
        Lazy loading of card content from json file.

        """
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if self.json_path and not self.is_initialized:
                self.is_initialized = True
                card_dict = load_json(self.json_path)
                self._anki_id = card_dict.get('anki_id')
                self._source_references = card_dict.get('source_references', {})
                self.index = card_dict.get('index')

                for name, field in card_dict.get('fields', {}).items():
                    super(Card, self).__setitem__(name, CardSection.from_dict(field))
            return method(self, *args, **kwargs)
        return wrapper

    @ensure_initialized
    def items(self):
        return super().items()

    @ensure_initialized
    def values(self):
        return super().values()

    @ensure_initialized
    def __getitem__(self, key):
        return super().__getitem__(key)

    @ensure_initialized
    def __setitem__(self, key, value):
        if not self.was_modified:
            self.was_modified = True
        return super().__setitem__(key, value)

    @property
    @ensure_initialized
    def source_references(self):
        return self._source_references

    @source_references.setter
    def source_references(self, value):
        print(f'Modifying sources from {self._source_references} to {value}')
        self._source_references = value
        if not self.was_modified:
            self.was_modified = True

    @property
    @ensure_initialized
    def anki_id(self):
        return self._anki_id

    @anki_id.setter
    def anki_id(self, value):
        print(f'Modifying anki_id from {self._anki_id} to {value}')
        self._anki_id = value
        if not self.was_modified:
            self.was_modified = True

    @property
    @ensure_initialized
    def frequency(self):
        """
        total frequency of word aggregated over all sources
        """
        return sum([freq for freq in self.source_references.values()])

    def generate_front(self):
        return f'<p style="font-size: 30px;">{self.word}</p>'

    @ensure_initialized
    def generate_back(self, field_names):
        """
        Generate flashcard html from downloaded fields
        """
        with open(HTML_TEMPLATE_DIR / 'static.html', 'r') as f:
            static = f.read()
        css_files = HTML_TEMPLATE_DIR.glob('*.css')
        css = '\n'.join(css_file.read_text(encoding='utf-8') for css_file in css_files)
        js_files = HTML_TEMPLATE_DIR.glob('*.js')
        js = '\n'.join(js_files.read_text(encoding='utf-8') for js_files in js_files)
        with open(HTML_TEMPLATE_DIR / 'flashcard.html', 'r') as f:
            template = Template(f.read())
        back = template.render(
            static_content=static,
            css_content=css,
            js_content=js,
            flashcard_content='<br>'.join([str(self[field_name]) for field_name in field_names])
        )
        return back

    @staticmethod
    def from_dict(word, json_path):
        """
        Create card object from dict
        Relevant when loading cards from nosql database or for testing
        """
        card = Card(word, json_path)
        return card

    @ensure_initialized
    def as_dict(self):
        """
        Dump card info as dict
        Relevant when storing nosql database and for printing
        """
        return {
            'source_references': self.source_references,
            'fields': {field_name: field.as_dict() for field_name, field in self.items()},
            'anki_id': self.anki_id,
            'index': self.index
        }

    @ensure_initialized
    def __repr__(self):
        return f'word: {self.word}\n' + \
               'occurrence frequencies in decks:\n' + \
               '\n'.join([f'    {source_name}: {freq}' for source_name, freq in self.source_references.items()]) + '\n' + \
               'fields with content: \n' + \
               '\n'.join([f'    {field_name}' for field_name in self.keys()]) + '\n' + \
               f'anki_id: {self.anki_id}'



from pathlib import Path
from collections import OrderedDict, Counter
from copy import deepcopy
import requests
import logging
import shutil
from collections.abc import Iterable
import json
import fnmatch
from urllib3.exceptions import MaxRetryError

from typing import Union, Dict, Iterable, List, Any

from . import PROJECT_ROOT, DATA_DIR, ANKIPAN_DB_ADDR
from . import Scraper
from . import AnkiManager
from . import Reader, File
from . import Card
from . import Deck
from . import util
from . import CardSection
from . import load_json, save_json

logger = logging.getLogger(__name__)

class Collection:
    def __init__(self,
                 name: str,
                 *,
                 source_lang: str = None,
                 native_lang: str = None,
                 data_dir = DATA_DIR):
        """
        Load new or existing collection.
        A collection can hold 0 to n decks.
        A deck can hold 1 to n flashcards.
        Flashcards have 1 to n fields, which refers to information on the backside (dictionary definitions, example sentences etc.)

        Parameters
        ----------
        name : Name of collection.
            Creates new collection for new names, loads existing collection for existing names.
        source_lang : Name of language the user wants to learn.
            Currently available: ['jp', 'de', 'en', 'fr']
        native_lang : Native language of the user for translations and explanations.
            Currently available: ['en', 'de']

        """
        self.name = name
        self.data_dir = data_dir
        self.metadata_path = data_dir / name / 'metadata.json'
        self.flashcard_fields_dir = data_dir / name / 'flashcard_fields'
        metadata = load_json(self.metadata_path)
        if metadata:
            if source_lang or native_lang:
                raise RuntimeError(f'Initializing existing database "{name}", specified kwargs would be ignored')
            self.source_lang = metadata.get('source_lang', source_lang)
            self.native_lang = metadata.get('native_lang', native_lang)
            self.source_names = metadata.get('source_names', [])
            self.flashcard_fields = metadata.get('flashcard_fields', {})
            self.known_sources = metadata.get('known_sources', {})
            self.cards = {}
            for word in metadata.get('card_words', []):
                self.cards[word] = Card.from_dict(word, self.flashcard_fields_dir / f'{word}.json')
            self.known_words = {}
            for known_word in metadata.get('known_words', []):
                self.known_words[known_word] = Card.from_dict(word, self.flashcard_fields_dir / f'{word}.json')
        else:
            if not (source_lang and native_lang):
                raise RuntimeError('source_lang and native_lang kwargs required when initializing new collection')
            self.source_lang = source_lang
            self.native_lang = native_lang
            self.source_names = []
            self.flashcard_fields = {}
            self.known_sources = {}
            self.cards = {}
            self.known_words = {}

        self.anki_manager = AnkiManager()
        self.reader = Reader(self.source_lang)
        self.scraper = Scraper(self.source_lang, self.native_lang)
        self._available_sources = None
        self.database_unavailable = False
        self.valid_definition_fields = [func.removeprefix('fetch_') for func in
                                fnmatch.filter([func for func in dir(Scraper)
                                                if callable(getattr(Scraper, func))], 'fetch_*')]


    def set_flashcard_fields(self,
                             definitions: List[str] = None,
                             sentence_sources: List[str] = None):
        """
        Tracks categories of sources and
        concrete sources that the user is familiar with.

        Examples
        --------
        collection = Collection('testcollection', source_lang='jp', native_lang='en')
        collection.get_available_sources()
            --> {'anime': ['One Piece', 'Bleach', ...], 'lyrics': {'Eve': [...], ...}, ...}
        # Get Word percentile information in word overview and display example sentences from categories on flashcards
        # Prioritize and highlight familiar sources in particular (Always try to include One piece example sentences if possible)
        # Also include other sources if not enough example sentences are available

        OR

        collection.get_available_sources().keys()
            --> dict_keys(['Netflix', 'anime', 'jpsubbers', 'lyrics', 'misc', 'youtube'])
            --> {'anime': [], 'lyrics': [], 'youtube': [], }
        # Still generates percentile information and displays generic example sentences from categories

        """
        invalid_fields = []
        assert isinstance(definitions, Iterable)
        assert isinstance(sentence_sources, Iterable)
        if definitions:
            for definition in definitions:
                if definition not in self.valid_definition_fields:
                    invalid_fields.append(definition)
        if sentence_sources:
            available_sources = self.get_available_sources()
            if available_sources is None:
                raise RuntimeError('Server currently unavailable, setting sentence sources not possible')
            for sentence_source in sentence_sources:
                if sentence_source not in available_sources.keys():
                    invalid_fields.append(sentence_source)
        if invalid_fields:
            raise RuntimeError(
                f"Some of the specified fields are invalid: {invalid_fields}\n\n"
                f"Valid definition field names (collection.valid_definition_fields):\n"
                f"{self.valid_definition_fields}\n\n"
                f"Valid sentence_source names (collection.get_available_sources()):\n"
                f"{list(available_sources.keys())}"
            )
        self.flashcard_fields['definitions'] = list(definitions) if definitions is not None else []
        self.flashcard_fields['sentence_sources'] = list(sentence_sources) if sentence_sources is not None else []

    @property
    def flashcard_fields_flattened(self):
        fields_flattened = []
        fields_flattened.extend([field for field in self.flashcard_fields['definitions']])
        fields_flattened.extend([f'sentences_{field}' for field in self.flashcard_fields['sentence_sources']])
        return fields_flattened

    @property
    def all_words(self):
        return self.cards | self.known_words

    @property
    def sorted_learning_words(self):
        return [word for word, card in sorted(self.cards.items(), key=lambda kv: kv[1].frequency, reverse=True)]

    @property
    def word_frequencies(self):
        return OrderedDict({word: card.frequency for word, card in sorted(self.cards.items(), key=lambda kv: kv[1].frequency, reverse=True)})

    @property
    def words_with_content(self):
        return [word for word in self.sorted_learning_words
                if all([field in self.cards[word].keys() for field in self.flashcard_fields_flattened])]

    def collect(self,
                paths: Union[Union[str, Path], Iterable[Union[str, Path]]] = None,
                *,
                source_name: str = None,
                string: str = None,
                word_counts: Dict[str, int] = None) -> Deck:
        """
        Collect words from specified source.

        Parameters
        ----------
        path: Path to a file or directory with textfiles to be added to the collection
            Uses stanza lemma parsing from reader.py to parse source files into wordcount dictionary

        kwargs (only consiered if path is not provided):
            source_name: Name of source in db to fetch lemmas from (does not require stanza)
            string: String to be parsed directly without reading a file
            word_counts: Dict of words and word_counts, directly adopted as Deck object

        """
        if not (isinstance(paths, list) or isinstance(paths, set)):
            if isinstance(paths, str) or isinstance(paths, Path):
                paths = [paths]
            elif not paths == None:
                raise RuntimeError(f'Unknown type passed as path: {paths}')
        deck = Deck(self.source_lang, self.native_lang,
                    learning_collection_words = set(self.cards.keys()),
                    known_collection_words = set(self.known_words.keys()),
                    sentence_sources = self.flashcard_fields.get('sentence_sources', []))
        if paths:
            for path in paths:
                deck.add(path=path)
        elif string:
            deck.add(string=string)
        elif word_counts:
            deck.add(word_counts=word_counts)
        elif source_name:
            data = {'lang': self.source_lang, 'source_name': source_name}
            res = requests.post(f'{ANKIPAN_DB_ADDR}/api/source_lemmas', json=data)
            if res.status_code == 404:
                raise RuntimeError(f'Ankipan DB 404 Error: {res.json()}')
            lemmas_and_count = res.json()
            deck.source_words.update(lemmas_and_count)
        return deck

    def add_deck(self,
            deck: Deck,
            source_name: str):
        """
        Add new deck from new words in Deck to current collection.
        Changes made to the new words and known words in the Deck object are adopted into the collection.

        Parameters
        ----------
        words: Words from Deck
        source_name: Name of source you are adding, e.g. movie or series title

        """

        if not source_name:
            raise RuntimeError('source_name is a mandatory field')
        if source_name in self.source_names:
            raise RuntimeError('source with same name already in collection')
        self.source_names.append(source_name)
        for word, freq in deck.known_words.items():
            if word in self.cards.keys(): #TODO: not needed?
                self.known_words[word] = self.cards.pop(word)
            elif word not in self.known_words:
                self.known_words[word] = Card(word, source_references = {source_name: freq})
            self.known_words[word].source_references[source_name] = freq
        for word, freq in deck.learning_words.items():
            self.cards[word].source_references[source_name] = freq
        for i, word in enumerate(deck.new_words):
            # TODO: Look into if this is too convoluted, we pass the available sources to the wordcollection, then get this info back? Maybe recalculate or store in db?
            self.cards[word] = Card(word, source_references = {source_name: deck.source_words[word]}, index = i)

    def remove(self,
               source_name: str):
        """
        Remove source added with specific name
                    file_paths = set()
        """
        # TODO: consider caching mechanism so downloaded info isn't destroyed
        if source_name not in self.source_names: raise RuntimeError(f'source name "{source_name}" not present in list of collection source names')
        self.source_names.remove(source_name)
        for stack in [self.cards, self.known_words]:
            superfluous_cards = []
            for word, card in stack.items():
                if source_name in card.source_references.keys():
                    if len(card.source_references) == 1:
                        superfluous_cards.append(word)
                    else:
                        stack[word].source_references.pop(source_name)
            [stack.pop(word) for word in superfluous_cards]

    # TODO: ignores existing sources and overwrites everything for some reason
    def fetch(self,
              deck_name: str,
              flashcard_fields: Iterable[str] = None,
              words: Iterable[str] = None,
              force_update = False):
        """
        Scrape/download data for new cards from sources

        Parameters
        ----------
        deck_name: Name of source to download data for
        flashcard_fields (optional): restrict fields to download data for
        words (optional): restrict words to download data for
        force_update: overwrite existing data

        """
        if not words:
            words = self.cards.keys()
        missing_fields = {}
        source_cards = [card for card in self.cards.values() if card.source_references and list(card.source_references.keys())[0] == deck_name and card.word in words]
        logger.debug(f'Fetching flashcard fields for "{deck_name}"')

        for card in source_cards:
            missing_fields_for_word = [field for field in (flashcard_fields if flashcard_fields else self.flashcard_fields_flattened)
                                         if not field in card.keys() or force_update]
            logger.debug(f'Missing fields for "{card.word}": {missing_fields_for_word}')
            for field in missing_fields_for_word:
                missing_fields.setdefault(field, []).append(card.word)
        if missing_fields:
            print(f'Fetching content for "{field}", {list(missing_fields.keys())}:')
            for field, words in missing_fields.items():
                print(f' - {field}: {words}')

            for i, (field, results) in enumerate(
                self.scraper.download_flashcard_fields(missing_fields,
                                                       known_sources_by_subdomain = self.known_sources).items()):
                for word in missing_fields[field]:
                    if word in results.keys():
                        self.cards[word][field] = results[word]
                    else:
                        self.cards[word][field] = CardSection('empty', 'black', '')
            self.save()

    def sync_with_anki(self, deck_name: str, overwrite = False):
        """
        Sync collection data with anki database.
        Uses AnkiConnect to interface with anki app functionalities via localhost.
        Requires anki to be installed, open and logged in during execution. (see README.md)
        https://apps.ankiweb.net/

        TODO: Currently slow at adding new cards, improve (windows issue)
        """

        # TODO: add "full_overwrite" kwarg, which bulk-removes and readds card instead of updating one by one (faster)
        source_cards = [card for card in self.cards.values() if deck_name in card.source_references.keys()]
        # TODO: sometimes not sorted? See Card.index creation
        source_cards.sort(key=lambda card: (card.index is None, card.index))
        incomplete_cards = {}
        for card in source_cards:
            for field in self.flashcard_fields_flattened:
                if field not in card.keys():
                    incomplete_cards.setdefault(card.word, []).append(field)
        if incomplete_cards:
            raise RuntimeError(f'Some cards are missing data for fields defined in Collection.flashcard_fields, run collection.fetch(<deck_name>) first: {incomplete_cards}')
        if source_cards:
            print('Syncing anki for words', [card.word for card in source_cards])
            self.anki_manager.sync_deck(deck_name, source_cards, self.flashcard_fields_flattened, overwrite=overwrite)
            self.anki_manager.sync()

    def save(self, name = None):
        """
        Write all collection data to local database.

        """
        if not name:
            name = self.name
        if not self.flashcard_fields_dir.exists():
            self.flashcard_fields_dir.mkdir(parents=True, exist_ok=True)
        print(f'saving collection "{name}"')
        metadata = {
            'source_lang': self.source_lang,
            'native_lang': self.native_lang,
            'flashcard_fields': self.flashcard_fields,
            'source_names': self.source_names,
            'card_words': list(self.cards.keys()),
            'known_words': list(self.known_words.keys()),
            'known_sources': self.known_sources
        }
        save_json(self.metadata_path, metadata)

        for word, card in {**self.cards, **self.known_words}.items():
            if card.was_modified:
                json_path = self.flashcard_fields_dir / f'{word}.json'
                save_json(json_path, card.as_dict())
                if not card.json_path:
                    card.json_path = json_path

    def estimate_known_words_for_domain(self, subdomain, level = None):
        word_counts = requests.post(f'{ANKIPAN_DB_ADDR}/api/corpus_wordcounts', json={'lang': self.source_lang, 'subdomains': [subdomain]}).json()
        proficiency_level = level*0.01*len(word_counts[subdomain]) if level else util.estimate_proficiency(word_counts[subdomain])
        items = Counter(word_counts[subdomain]).most_common()[:int(proficiency_level)]
        return self.collect(word_counts = dict(items), is_known = True)

    def get_available_sources(self) -> Dict[str, Dict[str, List[str]]]:
        if self.database_unavailable:
            logger.warning('Previous database connection attempt was unsuccessful, not reconnecting until Collection.database_unavailable is set to False again')
            return
        if self._available_sources is None:
            data = {'lang': self.source_lang}
            try:
                response = requests.post(f'{ANKIPAN_DB_ADDR}/api/available_sources', json=data)
                response.raise_for_status()  # Optional: ensure it's a 200-level response
                self._available_sources = response.json()
            except (requests.exceptions.JSONDecodeError,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.HTTPError,
                    requests.exceptions.RequestException) as e:
                logger.error(f'Server to fetch sources not available: {e}')
                self.database_unavailable = True
        return self._available_sources

    def add_custom(self, custom_explanations: Dict[str, CardSection]):
        """
        User method to add arbitrary flashcard fields to collection cards.

        Parameters
        ----------
        custom_explanations: dict
            Map of words to CardSection object
            Can also be simple string or html wrapped inside, see classes in scraper module

        """

        unknown_items = [key for key in custom_explanations if key not in self.all_words]
        if unknown_items:
            raise RuntimeError(f'Unknown items {unknown_items}')
        for word, explanation in custom_explanations.items():
            if word in self.all_words:
                self.all_words[word]['custom'] = explanation
            else:
                self.cards[word] = Card(word)
                self.cards[word]['custom'] = explanation

    @classmethod
    def load_from_dict(cls, name, dict: Union[str, Union[str, Iterable, Dict[str, Dict[str, Union[int, dict]]]]]) -> 'Collection':
        """
        Initialize collection from dictionary.

        Mostly used for testing.

        """
        dict_ = deepcopy(dict)
        card_words = dict_.pop('card_words')  if 'cards' in dict_.keys() else {}
        known_words = dict_.pop('known_words')  if 'known_words' in dict_.keys() else []
        source_names = dict_.pop('source_names') if 'source_names' in dict_.keys() else []

        instance = cls(name, **dict_)
        instance.cards = {word: Card.from_dict(word, PROJECT_ROOT / name / 'flashcard_fields' / f'{name}.json') for word in card_words}
        instance.known_words = {word: Card.from_dict(word, PROJECT_ROOT / name / 'flashcard_fields' / f'{name}.json') for word in known_words}
        instance.source_names = source_names

        return instance

    # TODO: Implement safety mechanism to not accidentally lose massive amounts of data
    # TODO: Implement function to recreate database based on current anki database state in case of loss
    def delete_collection(self, name):
        """
        Delete collection from database
        Currently ignores whether collection is present in database or not

        """
        if input('Are you sure? type "yes"') == 'yes':
            shutil.rmtree(self.data_dir / name)

    def as_dict(self):
        return {
            'source_lang': self.source_lang,
            'native_lang': self.native_lang,
            'flashcard_fields': self.flashcard_fields,
            'source_names': self.source_names,
            'card_words': [word for word in self.cards.keys()],
            'known_words': [word for word in self.cards.keys()],
            'known_sources': self.known_sources
        }

    def __repr__(self):
        return '\n'.join([
        f'name (collection name): {self.name}',
        f'source_lang: {self.source_lang}',
        f'native_lang: {self.native_lang}',
        f'source_names: {self.source_names}',
        f'flashcard_fields: {self.flashcard_fields}',
        f'known_sources: {self.known_sources}',
        f'n cards: {len(self.cards)}',
        f'n known_words: {len(self.known_words)}',
        f'n words_with_content: {len(self.words_with_content)}'])
